//#define AUDIO_LOG
/*
 * libdtmfr2 - a library for DTMF/R2 signaling on E1 lines
 *
 * dtmfr2.c
 *
 * Written by Steve Underwood <steveu@coppice.org>
 *
 * Copyright (C) 2002, 2003, 2005 Steve Underwood
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: dtmfr2.c,v 1.3 2005/12/10 14:48:48 steveu Exp $
 */

/*! \file libdtmfr2.h */

#define _ISOC9X_SOURCE  1
#define _ISOC99_SOURCE  1

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <limits.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include <zaptel/zaptel.h>

#include <tiffio.h>
#include <spandsp.h>

#include <unicall.h>
#include <unicall/hashtable.h>
#include <unicall/unicall-private.h>
#include <libsupertone.h>

#include "libdtmfr2.h"
#include "dtmfr2.h"

#define FALSE 0
#define TRUE (!FALSE)

#define BLOCKING_RELEASE_TIME               450
#define ANSWER_GUARD_TIME                   100

#define DEFAULT_MAX_SEIZE_ACK_WAIT          2000
#define DEFAULT_MAX_WAIT_FOR_GROUP_B_SIGNAL 15000
#define DEFAULT_MAX_AWAIT_ANSWER            60000

static uc_check_event_t check_event;
static uc_call_control_t call_control;
static uc_new_t create_new;
static uc_error_message_t error_message;
static uc_get_device_handles_t get_device_handles;
static uc_channel_open_t channel_open;
static uc_channel_close_t channel_close;
static uc_channel_set_api_codec_t channel_set_api_codec;
static uc_channel_write_t channel_write;
static uc_channel_flush_t channel_flush;
static uc_channel_gains_t channel_gains;
static uc_channel_echo_cancel_t channel_echo_cancel;
static uc_channel_switching_t channel_switching;

uc_protocol_t protocol_descriptor =
{
    "dtmfr2",
    "E1 DTMF/R2 signalling",
    "DTMF/R2",
    "\0",
    "\0",

    check_event,
    create_new,
    call_control,
    error_message,
    get_device_handles,
    channel_open,
    channel_close,
    channel_set_api_codec,
    channel_write,
    channel_flush,
    channel_gains,
    channel_echo_cancel,
    channel_switching
};

enum
{
    ACTIVE_NONE,
    ACTIVE_SPEECH,
    ACTIVE_DTMF,
    ACTIVE_SUPER_TONE
};

#define R2_STATUS_FORMAT "[%d/%8X/%-14s/%-13s]"

static void abcd_signaling_event(uc_t *uc, int ch);

static int set_abcd_signal(uc_t *uc, int ch, uint8_t abcd);

static void select_active_rxtx(uc_t *uc, int which);
static void seize_ack_wait_expired(uc_t *uc, void *data);
static void release_guard_expired(uc_t *uc, void *data);
static void far_unblocking_expired(uc_t *uc, void *data);
static void local_unblocking_expired(uc_t *uc, void *data);
static void answer_guard_expired(uc_t *uc, void *data);
static void clear_back_persistence_expired(uc_t *uc, void *data);
static void billing_pulse_expired(uc_t *uc, void *data);
static int send_calling_party_category_request(uc_t *uc);

static int variant_element(const char **variant, int which)
{
    int i;
    const char *s;
    const char *t;

    /* Find element "which" in a comma separated string of elements */
    t = *variant - 1;
    for (i = -1;  i < which;  i++)
    {
        s = t + 1;
        if ((t = strchr(t + 1, ',')) == NULL)
        {
            t = s + strlen(s);
            break;
        }
        /*endif*/
    }
    /*endfor*/
    if (i >= which - 1)
    {
        *variant = s;
        return t - s;
    }
    /*endif*/
    return 0;
}
/*- End of function --------------------------------------------------------*/

static int protocol_variant(const char *variant)
{
    int i;
    int inlen;
    int len;
    char const *s;
    char const *t;

    inlen = variant_element(&variant, 0);
    s = protocol_descriptor.variants;
    for (i = 1;  *s;  i++)
    {
        len = strlen(s);
        if (len == inlen  &&  strncmp(variant, s, len) == 0)
            return  i;
        /*endif*/
        s += (len + 1);
    }
    /*endfor*/
    return  -1;
}
/*- End of function --------------------------------------------------------*/

static char *r2_phase_to_str(int state)
{
    switch (state)
    {
    case 0:
        return "Idle";
    case R2_BACK_SEIZE_ACK:
        return "Seize ack";
    case R2_BACK_ANSWER:
        return "Answer";
    case R2_BACK_CLEAR_BACK:
        return "Clear back";
    case R2_BACK_CLEAR_FWD:
        return "Clear fwd";
    case R2_BACK_RELEASE_GUARD:
        return "Release guard";
    case R2_FWD_SEIZE:
        return "Seize";
    case R2_FWD_AWAIT_ANSWER:
        return "Await answer";
    case R2_FWD_ANSWERED:
        return "Answered";
    case R2_FWD_CLEAR_BACK:
        return "Clear back";
    case R2_FWD_CLEAR_FWD_BEFORE_SEIZE_ACK:
        return "Clear fwd A";
    case R2_FWD_CLEAR_FWD_BEFORE_ANSWER:
        return "Clear fwd B";
    case R2_FWD_CLEAR_FWD_AFTER_ANSWER:
        return "Clear fwd C";
    case R2_FWD_CLEAR_FWD_AFTER_CLEAR_BACK:
        return "Clear fwd D";
    }
    /*endswitch*/
    return "???";
}
/*- End of function --------------------------------------------------------*/

static char *r2_state_to_str(int state)
{
    switch (state)
    {
    case 0:
        return "Idle";
    case R2_SENT_SEIZE_ACK:
        return "Seize ack";
    case R2_SENT_ACCEPTED_CHARGE:
        return "Accepted Paid";
    case R2_SENT_ACCEPTED_NO_CHARGE:
        return "Accepted Free";
    case R2_SENT_NETWORK_CONGESTION:
        return "Congestion";
    case R2_SENT_UNASSIGNED_NUMBER:
        return "Unassigned";
    case R2_SENT_DEST_BUSY:
        return "Busy";
    case R2_SENT_DEST_OUT_OF_ORDER:
        return "Out of order";
    case R2_SENT_DNIS:
        return "DNIS";
    case R2_SENT_ANI:
        return "ANI";
    case R2_SENT_END_OF_DNIS:
        return "End of DNIS";
    case R2_SENT_END_OF_ANI:
        return "End of ANI";
    case R2_SENT_NOTHING:
        return "Silent";
    }
    /*endswitch*/
    return "???";
}
/*- End of function --------------------------------------------------------*/

static void handle_rx_dtmf(void *user_data, const char *digits, int len)
{
}
/*- End of function --------------------------------------------------------*/

static void select_active_rxtx(uc_t *uc, int which)
{
    dtmfr2_signaling_state_t *dtmfr2;
    int ch;
    
    ch = 0;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    /* The rx and tx active selections are ganged */
    uc->chan[ch].active_rx =
    uc->chan[ch].active_tx = which;
    if (which != ACTIVE_DTMF)
    {
        /* Any tone generation and detection should stop at any change */
    }
    /*endif*/
}
/*- End of function --------------------------------------------------------*/

static void protocol_error_recovery(uc_t *uc, int ch, int cause)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;

    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc,
           UC_LOG_PROTOCOL_ERROR,
           "R2 prot. err. " R2_STATUS_FORMAT " cause %d - %s\n",
           uc->chan[ch].active_rx,
           uc->chan[ch].state,
           r2_phase_to_str(dtmfr2->r2_phase),
           r2_state_to_str(dtmfr2->r2_state),
           cause,
           error_message(uc, cause));
    /* If we are blocked, a protocol problem isn't a good reason to unblock.
       However, for other starting conditions, going to the idle state seems OK. */
    if (uc->chan[ch].state != UC_STATE_BLOCKED)
        uc->chan[ch].state = UC_STATE_IDLE;
    /*endif*/

    /* Settle everything back to its initial state */
    dtmfr2->r2_phase = 0;
    dtmfr2->r2_state = 0;
    dtmfr2->calling_party_category_exchanged = FALSE;
    select_active_rxtx(uc, ACTIVE_SPEECH);
    set_abcd_signal(uc, ch, dtmfr2->abcd_idle);
    set_mf_signal(uc, ch, 0);

    if (dtmfr2->far_unblocking_timer >= 0)
    {
        uc_schedule_del(uc, dtmfr2->far_unblocking_timer);
        dtmfr2->far_unblocking_timer = -1;
    }
    /*endif*/
    if (dtmfr2->local_unblocking_timer >= 0)
    {
        uc_schedule_del(uc, dtmfr2->local_unblocking_timer);
        dtmfr2->local_unblocking_timer = -1;
    }
    /*endif*/
    if (dtmfr2->answer_guard_timer >= 0)
    {
        uc_schedule_del(uc, dtmfr2->answer_guard_timer);
        dtmfr2->answer_guard_timer = -1;
    }
    /*endif*/
    if (dtmfr2->await_answer_timer >= 0)
    {
        uc_schedule_del(uc, dtmfr2->await_answer_timer);
        dtmfr2->await_answer_timer = -1;
    }
    /*endif*/
    if (dtmfr2->release_guard_timer >= 0)
    {
        uc_schedule_del(uc, dtmfr2->release_guard_timer);
        dtmfr2->release_guard_timer = -1;
    }
    /*endif*/
    if (dtmfr2->clear_back_persistence_timer >= 0)
    {
        uc_schedule_del(uc, dtmfr2->clear_back_persistence_timer);
        dtmfr2->clear_back_persistence_timer = -1;
    }
    /*endif*/
    if (dtmfr2->billing_pulse_timer >= 0)
    {
        uc_schedule_del(uc, dtmfr2->billing_pulse_timer);
        dtmfr2->billing_pulse_timer = -1;
    }
    /*endif*/

    ev.e = UC_EVENT_PROTOCOLFAIL;
    ev.gen.channel = ch;
    ev.gen.data = cause;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static int set_abcd_signal(uc_t *uc, int ch, uint8_t abcd)
{
    dtmfr2_signaling_state_t *dtmfr2;
    int x;

    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    abcd |= dtmfr2->base_abcd_bits;
    uc_log(uc,
           UC_LOG_DEBUG_2,
           "%d%d%d%d  ->      " R2_STATUS_FORMAT "\n",
           (abcd >> 3) & 1, (abcd >> 2) & 1, (abcd >> 1) & 1, abcd & 1,
           uc->chan[ch].active_rx,
           uc->chan[ch].state,
           r2_phase_to_str(dtmfr2->r2_phase),
           r2_state_to_str(dtmfr2->r2_state));
    x = abcd;
    if (ioctl(uc->chan[ch].handle, ZT_SETTXBITS, &x) < 0)
    {
        protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_SET_CAS_FAILURE);
        return  -1;
    }
    /*endif*/
    return  0;
}
/*- End of function --------------------------------------------------------*/

static void start_detected(uc_t *uc, int ch)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;
    uc_call_t *call;

    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Detected\n");

    /* TODO: a memory allocation issue at this point is not handled well. */
    if ((call = uc_createcall(uc, 0)) == NULL)
    {
        uc_log(uc, UC_LOG_ERROR, "Failed to allocate a new call structure\n");
        return;
    }
    /*endif*/
    dtmfr2->call = call;

    call->incoming_call = TRUE;

    /* Right now we have no received call information. */
    dtmfr2->ani_cnt = 0;
    dtmfr2->ani_len = 0;
    dtmfr2->dnis_cnt = 0;
    dtmfr2->dnis_len = 0;
    dtmfr2->dnis_complete = FALSE;
    dtmfr2->calling_party_category_exchanged = FALSE;

    call->chan = ch;
    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_DETECTED;
    /*endif*/
    call->state = UC_STATE_DETECTED;
    /* Set a default for the final disconnection */
    call->disconnect_reason = UC_CAUSE_UNSPECIFIED_CAUSE;
    /* We need to choose the codec used by the card. */
    call->callparms.userinfo_layer1_protocol = UC_LAYER_1_ALAW;

    dtmf_rx_init(&dtmfr2->dtmf_rx_state, handle_rx_dtmf, dtmfr2);
    select_active_rxtx(uc, ACTIVE_DTMF);
    /* The seize ack tells the other end our DTMF detector is
       active, and tone signaling may begin. */
    if (set_abcd_signal(uc, ch, dtmfr2->back_abcd_seize_ack))
        return;
    /*endif*/

    dtmfr2->r2_phase = R2_BACK_SEIZE_ACK;
    dtmfr2->r2_state = R2_SENT_SEIZE_ACK;

    ev.e = UC_EVENT_DETECTED;
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static void start_call_disconnected(uc_t *uc, uc_call_t *call)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;

    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc,
           UC_LOG_FLOW,
           "Call disconnected(cause=%s [%d]) - state 0x%X\n",
           uc_cause2str(call->disconnect_reason),
           call->disconnect_reason,
           call->state);
    
    /* Make sure any tone processing is stopped, if we disconnected during
       call setup. */
    select_active_rxtx(uc, ACTIVE_SPEECH);

    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_CALL_DISCONNECTED;
    /*endif*/
    call->state = UC_STATE_CALL_DISCONNECTED;

    ev.fardisconnected.e = UC_EVENT_DROPCALL;
    ev.fardisconnected.channel = call->chan;
    ev.fardisconnected.cause = call->disconnect_reason;
    ev.fardisconnected.crn = call->crn;
    ev.fardisconnected.call = call;
    uc_report_event(uc, &ev, sizeof(ev));
}
/*- End of function --------------------------------------------------------*/

static void start_far_disconnected(uc_t *uc, uc_call_t *call)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;

    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc,
           UC_LOG_FLOW,
           "Far end disconnected(cause=%s [%d]) - state 0x%X\n",
           uc_cause2str(call->disconnect_reason),
           call->disconnect_reason,
           call->state);

    /* Don't stop the tone detector here. */

    if (dtmfr2->await_answer_timer >= 0)
    {
        uc_schedule_del(uc, dtmfr2->await_answer_timer);
        dtmfr2->await_answer_timer = -1;
    }
    /*endif*/
    if (dtmfr2->answer_guard_timer >= 0)
    {
        uc_schedule_del(uc, dtmfr2->answer_guard_timer);
        dtmfr2->answer_guard_timer = -1;
    }
    /*endif*/
    if (dtmfr2->clear_back_persistence_timer >= 0)
    {
        uc_schedule_del(uc, dtmfr2->clear_back_persistence_timer);
        dtmfr2->clear_back_persistence_timer = -1;
    }
    /*endif*/
    if (dtmfr2->billing_pulse_timer >= 0)
    {
        uc_schedule_del(uc, dtmfr2->billing_pulse_timer);
        dtmfr2->billing_pulse_timer = -1;
    }
    /*endif*/
    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_FAR_DISCONNECTED;
    /*endif*/
    call->state = UC_STATE_FAR_DISCONNECTED;

    ev.fardisconnected.e = UC_EVENT_FARDISCONNECTED;
    ev.fardisconnected.channel = call->chan;
    ev.fardisconnected.cause = call->disconnect_reason;
    ev.fardisconnected.crn = call->crn;
    ev.fardisconnected.call = call;
    uc_report_event(uc, &ev, sizeof(ev));
}
/*- End of function --------------------------------------------------------*/

static void start_idle(uc_t *uc, uc_call_t *call)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;

    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    /* N.B. This is really only for starting the idle state from a call. From
       initialisation, unblocking, or error recovery this is not the right
       choice. */
    /* Make sure no junk was left behind. */
    dtmfr2->r2_phase = 0;
    dtmfr2->r2_state = 0;
    dtmfr2->calling_party_category_exchanged = FALSE;
    select_active_rxtx(uc, ACTIVE_SPEECH);

    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_IDLE;
    /*endif*/
    call->state = UC_STATE_IDLE;

    ev.e = UC_EVENT_RELEASECALL;
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));

    /* Although we destroy the call here, the event sent just above may not
       have been delivered. It is important, therefore, that the application
       only treat the call as a reference tag, and does not try to look at
       its contents. */
    uc_destroycall(uc, call->crn);
    dtmfr2->call = NULL;
}
/*- End of function --------------------------------------------------------*/

static void start_offered(uc_t *uc, uc_call_t *call)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;

    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_OFFERED;
    /*endif*/
    call->state = UC_STATE_OFFERED;
    
    call->disconnect_reason = UC_CAUSE_NORMAL_CLEARING;

    memset (&ev.offered, '\0', sizeof(ev.offered));
    ev.offered.e = UC_EVENT_OFFERED;
    ev.offered.channel = call->chan;
    strcpy (ev.offered.parms.originating_number, call->callparms.originating_number);
    strcpy (ev.offered.parms.destination_number, call->callparms.destination_number);
    ev.offered.parms.calling_party_category = call->callparms.calling_party_category;
    ev.offered.parms.call_type = call->callparms.call_type;
    ev.offered.parms.userinfo_layer1_protocol = call->callparms.userinfo_layer1_protocol;
    ev.offered.crn = call->crn;
    ev.offered.call = call;
    uc_report_event(uc, &ev, sizeof(ev));
}
/*- End of function --------------------------------------------------------*/

static void start_alerting(uc_t *uc, uc_call_t *call)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;

    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    select_active_rxtx(uc, ACTIVE_SPEECH);
    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_ALERTING;
    /*endif*/
    call->state = UC_STATE_ALERTING;

    ev.e = UC_EVENT_ALERTING;
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static void start_connected(uc_t *uc, uc_call_t *call)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;

    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_CONNECTED;
    /*endif*/
    call->state = UC_STATE_CONNECTED;
    
    select_active_rxtx(uc, ACTIVE_SPEECH);
    /* Set the default for the clear that will occur later */
    call->disconnect_reason = UC_CAUSE_NORMAL_CLEARING;

    if (call->incoming_call)
    {
        ev.e = UC_EVENT_ANSWERED;
    }
    else
    {
        if (dtmfr2->await_answer_timer >= 0)
        {
            uc_schedule_del(uc, dtmfr2->await_answer_timer);
            dtmfr2->await_answer_timer = -1;
        }
        /*endif*/
        ev.e = UC_EVENT_CONNECTED;
    }
    /*endif*/
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static int start_clear_fwd(uc_t *uc)
{
    dtmfr2_signaling_state_t *dtmfr2;
    int ch;

    ch = 0;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Clearing fwd\n");

    if (set_abcd_signal(uc, ch, dtmfr2->fwd_abcd_clear_fwd))
        return  UC_RET_BAD_DEVICE;
    /*endif*/
    dtmfr2->r2_state = 0;
    dtmfr2->calling_party_category_exchanged = FALSE;
    select_active_rxtx(uc, ACTIVE_SPEECH);
    return UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static void seize_ack_wait_expired(uc_t *uc, void *data)
{
    dtmfr2_signaling_state_t *dtmfr2;
    int ch;

    ch = (intptr_t) data;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "seize_ack_wait_expired\n");
    dtmfr2->seize_ack_timer = -1;
    protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_SEIZE_ACK_TIMEOUT);
}
/*- End of function --------------------------------------------------------*/

static void release_guard_expired(uc_t *uc, void *data)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_call_t *call;

    call = (uc_call_t *) data;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Release guard expired\n");
    dtmfr2->release_guard_timer = -1;
    start_idle(uc, call);
}
/*- End of function --------------------------------------------------------*/

static void answer_guard_expired(uc_t *uc, void *data)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;
    uc_call_t *call;
    int ch;

    call = (uc_call_t *) data;
    ch = 0;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Answer guard expired\n");
    /* The answer guard time ensures we do not answer with the CAS bits before
       the audio buffer has flushed tone out to the far end. Some MFC/R2
       kit doesn't like this kind of out of sequence answering. */
    /* TODO: Waiting for the tone to drain from the write buffer by the buffer
             empty event would be crisper. Adding some margin to the end of that
             might make things more tolerant, though. */
    dtmfr2->answer_guard_timer = -1;
    /* If answer occured before the timeout, we apply the answer condition now. */
    if (call->state == UC_STATE_CONNECTED)
    {
        if (set_abcd_signal(uc, ch, dtmfr2->back_abcd_answer))
            return;
        /*endif*/
        dtmfr2->r2_phase = R2_BACK_ANSWER;
        start_connected(uc, dtmfr2->call);
    }
    else
    {
        if (call->chan >= 0)
            uc->chan[call->chan].state = UC_STATE_ACCEPTED;
        /*endif*/
        call->state = UC_STATE_ACCEPTED;

        ev.e = UC_EVENT_ACCEPTED;
        ev.gen.channel = ch;
        ev.gen.crn = call->crn;
        ev.gen.call = call;
        uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
    }
    /*endif*/
}
/*- End of function --------------------------------------------------------*/

static void clear_back_persistence_expired(uc_t *uc, void *data)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_call_t *call;
    int ch;
    
    call = (uc_call_t *) data;
    ch = 0;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    dtmfr2->clear_back_persistence_timer = -1;
    start_far_disconnected(uc, dtmfr2->call);
}
/*- End of function --------------------------------------------------------*/

static void billing_pulse_expired(uc_t *uc, void *data)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_call_t *call;
    int ch;
    
    call = (uc_call_t *) data;
    ch = 0;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    dtmfr2->billing_pulse_timer = -1;
    /* Return to the answer state, if we are still connected. */
    if (call->state == UC_STATE_CONNECTED)
        set_abcd_signal(uc, ch, dtmfr2->back_abcd_answer);
    /*endif*/
}
/*- End of function --------------------------------------------------------*/

static void far_unblocking_expired(uc_t *uc, void *data)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;
    int ch;
    
    ch = (intptr_t) data;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "far_unblocking_expired\n");

    dtmfr2->far_unblocking_timer = -1;

    dtmfr2->far_blocked = FALSE;
    if (uc->chan[ch].state == UC_STATE_BLOCKED)
        uc->chan[ch].state = UC_STATE_IDLE;
    /*endif*/

    ev.e = UC_EVENT_FARUNBLOCKED;
    ev.gen.channel = ch;
    ev.gen.crn = 0;
    ev.gen.call = NULL;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static void local_unblocking_expired(uc_t *uc, void *data)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;
    int ch;
    
    ch = (intptr_t) data;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "local_unblocking_expired\n");

    dtmfr2->local_unblocking_timer = -1;

    dtmfr2->local_blocked = FALSE;

    ev.e = UC_EVENT_LOCALUNBLOCKED;
    ev.gen.channel = ch;
    ev.gen.crn = 0;
    ev.gen.call = NULL;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static void await_answer_expired(uc_t *uc, void *data)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_call_t *call;
    
    call = (uc_call_t *) data;    
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Timed out waiting for answer\n");

    dtmfr2->await_answer_timer = -1;

    /* Treat this as though the other end cleared the call. */
    call->disconnect_reason = UC_CAUSE_NO_ANSWER_FROM_USER;
    start_far_disconnected(uc, call);
}
/*- End of function --------------------------------------------------------*/

static void abcd_signaling_event(uc_t *uc, int ch)
{
    int abcd;
    int masked_abcd;
    int changes;
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;
    
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    if (ioctl(uc->chan[ch].handle, ZT_GETRXBITS, &abcd))
    {
        uc_log(uc, UC_LOG_ERROR, "Failed to get CAS bits - %d\n", errno);
        return;
    }
    /*endif*/
    uc_log(uc,
           UC_LOG_DEBUG_2,
           "     <- %d%d%d%d  " R2_STATUS_FORMAT "\n",
           (abcd >> 3) & 1, (abcd >> 2) & 1, (abcd >> 1) & 1, abcd & 1,
           uc->chan[ch].active_rx,
           uc->chan[ch].state,
           r2_phase_to_str(dtmfr2->r2_phase),
           r2_state_to_str(dtmfr2->r2_state));
    changes = dtmfr2->current_abcd_in ^ abcd;
    dtmfr2->current_abcd_in = abcd;
    /* TODO: deal with charging pulses, and other things, which occur in the C and D
             bits in some places. */
    /* TODO: deal with charging pulses, which occur on the A or B bits during the connected
             phase in some places. */
    /* The bits which changed might not be the R2 signaling bits. */
    if ((changes & dtmfr2->abcd_mask) == 0)
        return;
    /*endif*/
    masked_abcd = abcd & dtmfr2->abcd_mask;
    switch (uc->chan[ch].state)
    {
    case UC_STATE_NULL:
        /* In reality, for polling mode, this branch is never taken. */
        if (masked_abcd == dtmfr2->abcd_idle)
        {
            uc_log(uc, UC_LOG_WARNING, "Got CAS idle in UC_STATE_NULL\n");
            uc->chan[ch].state = UC_STATE_IDLE;
            dtmfr2->r2_phase = 0;
            dtmfr2->r2_state = 0;
            dtmfr2->calling_party_category_exchanged = FALSE;
            select_active_rxtx(uc, ACTIVE_SPEECH);
        }
        else
        {
            /* In initial state, any signal could be received. */
        }
        /*endif*/
        return;
    case UC_STATE_IDLE:
        if (masked_abcd == dtmfr2->abcd_idle)
        {
            if (set_abcd_signal(uc, ch, dtmfr2->abcd_idle))
                return;
            /*endif*/
        }
        else if (masked_abcd == dtmfr2->fwd_abcd_seize)
        {
            start_detected(uc, ch);
        }
        else if (masked_abcd == dtmfr2->abcd_blocking)
        {
            uc->chan[ch].state = UC_STATE_BLOCKED;
            dtmfr2->far_blocked = TRUE;
            select_active_rxtx(uc, ACTIVE_SPEECH);
            ev.e = UC_EVENT_FARBLOCKED;
            ev.gen.channel = ch;
            uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
        }
        else
        {
            protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
        }
        /*endif*/
        return;
    case UC_STATE_BLOCKED:
        if (masked_abcd == dtmfr2->abcd_idle)
        {
            /* Start a timeout to ensure we do not seize an unblocked trunk too
               agressively. */
            dtmfr2->far_unblocking_timer =
                uc_schedule_event(uc, dtmfr2->blocking_release_time, far_unblocking_expired, (void *) (intptr_t) ch);
        }
        else if (masked_abcd == dtmfr2->fwd_abcd_seize)
        {
            /* We must have got a seize before we expired an unblocking timer.
               Force the end of the blocking timer, and accept the call. */
            if (dtmfr2->far_unblocking_timer >= 0)
            {
                uc_schedule_del(uc, dtmfr2->far_unblocking_timer);
                dtmfr2->far_unblocking_timer = -1;
                far_unblocking_expired(uc, (void *) (intptr_t) ch);
            }
            /*endif*/
            if (dtmfr2->local_unblocking_timer >= 0)
            {
                uc_schedule_del(uc, dtmfr2->local_unblocking_timer);
                dtmfr2->local_unblocking_timer = -1;
                local_unblocking_expired(uc, (void *) (intptr_t) ch);
            }
            /*endif*/
            start_detected(uc, ch);
        }
        else if (masked_abcd == dtmfr2->abcd_blocking)
        {
            /* This could happen if the line is blocked at startup */
            uc->chan[ch].state = UC_STATE_BLOCKED;
            dtmfr2->far_blocked = TRUE;
            select_active_rxtx(uc, ACTIVE_SPEECH);
            ev.e = UC_EVENT_FARBLOCKED;
            ev.gen.channel = ch;
            uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
        }
        else
        {
            protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
        }
        /*endif*/
        return;
    }
    /*endswitch*/

    /* There must be a call in progress */
    if (dtmfr2->call == NULL)
    {
        /* This is bad. Try to avoid things falling apart, by doing a protocol recovery.
           However, if we are here it is not a protocol problem. Its a bug. */
        uc_log(uc, UC_LOG_ERROR, "!!!! ABCD in call state 0x%X with no call structure\n", uc->chan[ch].state);
        protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_INVALID_STATE);
        return;
    }
    /*endif*/
    if (dtmfr2->call->incoming_call)
    {
        switch (dtmfr2->r2_phase)
        {
        case R2_BACK_SEIZE_ACK:
        case R2_BACK_DTMF:
        case R2_BACK_ANSWER:
            if (masked_abcd == dtmfr2->fwd_abcd_clear_fwd)
            {
                dtmfr2->r2_phase = R2_BACK_CLEAR_FWD;
                start_far_disconnected(uc, dtmfr2->call);
            }
            else
            {
                protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            }
            /*endif*/
            break;
        case R2_BACK_CLEAR_BACK:
            if (masked_abcd == dtmfr2->fwd_abcd_clear_fwd)
                start_call_disconnected(uc, dtmfr2->call);
            else
                protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            /*endif*/
            break;
        case R2_BACK_RELEASE_GUARD:
            if (masked_abcd == dtmfr2->fwd_abcd_seize)
            {
                /* An enthusiastic far end has seized for a new call, while we
                   were still timing out the last one. Make like the timeout
                   occured, and carry on with the new call. */
                if (dtmfr2->release_guard_timer >= 0)
                    uc_schedule_del(uc, dtmfr2->release_guard_timer);
                /*endif*/
                release_guard_expired(uc, (void *) dtmfr2->call);

                /* .... and now the new call begins. */
                start_detected(uc, ch);
            }
            else
            {
                protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            }
            /*endif*/
            break;
        default:
            /* Unknown phase. Handle as a protocol error. */
            protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_INVALID_STATE);
            break;
        }
        /*endswitch*/
    }
    else
    {
        switch (dtmfr2->r2_phase)
        {
        case R2_FWD_SEIZE:
            if (masked_abcd == dtmfr2->back_abcd_seize_ack)
            {
                if (dtmfr2->seize_ack_timer >= 0)
                {
                    uc_schedule_del(uc, dtmfr2->seize_ack_timer);
                    dtmfr2->seize_ack_timer = -1;
                }
                /*endif*/
                /* The other side's DTMF detector should now be listening.
                   We have not started sending DTMF, so we cannot have missed
                   any incoming DTMF if we only enable the detector now. Enabling
                   the detector runs the risk of picking up something from the
                   speech path, before the other end muted it. MF6 is not very
                   speech immune. */
                dtmf_tx_init(&dtmfr2->dtmf_tx_state);
                select_active_rxtx(uc, ACTIVE_DTMF);
                /* Start the MF phase by sending the DNIS. The other
                   end may change from DNIS by the signals it returns,
                   but we always start with the DNIS. */
                dtmfr2->r2_phase = R2_FWD_DTMF;
                dtmfr2->dnis_cnt = 0;
                dtmfr2->ani_cnt = 0;
                send_dnis_digit(uc, ch, 0);
            }
            else if (masked_abcd == dtmfr2->fwd_abcd_seize)
            {
                /* Call collision! We should clear, and abandon this call attempt. */
                dtmfr2->call->disconnect_reason = UC_CAUSE_NETWORK_CONGESTION;
                /* Provoke the application to drop the call. */
                start_far_disconnected(uc, dtmfr2->call);
            }
            else
            {
                protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            }
            /*endif*/
            break;
        case R2_FWD_DTMF:
            if (masked_abcd == dtmfr2->back_abcd_answer)
            {
                /* Timing overlaps might legitimately put us in this
                   position. */
                dtmfr2->r2_phase = R2_FWD_ANSWERED_BEFORE_GROUP_B_SIGNAL;
            }
            else if (masked_abcd == dtmfr2->back_abcd_clear_back)
            {
                dtmfr2->r2_phase = R2_FWD_CLEAR_BACK;
                start_far_disconnected(uc, dtmfr2->call);
            }
            else
            {
                protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            }
            /*endif*/
            break;
        case R2_FWD_AWAIT_ANSWER:
            if (masked_abcd == dtmfr2->back_abcd_answer)
            {
                dtmfr2->r2_phase = R2_FWD_ANSWERED;
                start_connected(uc, dtmfr2->call);
            }
            else if (masked_abcd == dtmfr2->back_abcd_clear_back)
            {
                dtmfr2->r2_phase = R2_FWD_CLEAR_BACK;
                start_far_disconnected(uc, dtmfr2->call);
            }
            else
            {
                protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            }
            /*endif*/
            break;
        case R2_FWD_ANSWERED:
            if (masked_abcd == dtmfr2->back_abcd_clear_back)
            {
                dtmfr2->r2_phase = R2_FWD_CLEAR_BACK;
                if (dtmfr2->clear_back_persistence_check)
                {
                    /* We need to apply a persistence check to separate real clearbacks
                       from metering pulses. */
                    dtmfr2->clear_back_persistence_timer =
                        uc_schedule_event(uc, dtmfr2->clear_back_persistence_check, clear_back_persistence_expired, dtmfr2->call);
                }
                else
                {
                    start_far_disconnected(uc, dtmfr2->call);
                }
                /*endif*/
            }
            else
            {
                protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            }
            /*endif*/
            break;
        case R2_FWD_CLEAR_BACK:
            if (dtmfr2->clear_back_persistence_timer >= 0)
            {
                /* We have not yet confirmed the clear back condition. */
                uc_schedule_del(uc, dtmfr2->clear_back_persistence_timer);
                dtmfr2->clear_back_persistence_timer = -1;
                if (masked_abcd == dtmfr2->back_abcd_answer)
                {
                    /* It must have been a metering pulse. Return to the answered state. */
                    dtmfr2->r2_phase = R2_FWD_ANSWERED;
                    break;
                }
                /*endif*/
            }
            /*endif*/
            /* If they applied CLEAR BACK, there should be no more CAS transitions
               until we take some action. */
            protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            break;
        case R2_FWD_CLEAR_FWD_BEFORE_SEIZE_ACK:
            if (masked_abcd == dtmfr2->back_abcd_seize_ack)
            {
                if (dtmfr2->seize_ack_timer >= 0)
                {
                    uc_schedule_del(uc, dtmfr2->seize_ack_timer);
                    dtmfr2->seize_ack_timer = -1;
                }
                /*endif*/
                /* It is now safe to send the pending CLEAR FWD. */
                if (start_clear_fwd(uc))
                    return;
                /*endif*/
            }
            else if (masked_abcd == dtmfr2->fwd_abcd_seize)
            {
                /* Call collision! We should clear, and abandon this call attempt. */
                dtmfr2->call->disconnect_reason = UC_CAUSE_NETWORK_CONGESTION;
                if (start_clear_fwd(uc))
                    return;
                /*endif*/
                /* The application has already dropped this call, so just wait
                   until the far end reacts and disconnects. */
            }
            else
            {
                protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            }
            /*endif*/
            break;
        case R2_FWD_CLEAR_FWD_BEFORE_ANSWER:
            if (masked_abcd == dtmfr2->abcd_idle)
                start_idle(uc, dtmfr2->call);
            else
                protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            /*endif*/
            break;
        case R2_FWD_CLEAR_FWD_AFTER_ANSWER:
            if (masked_abcd == dtmfr2->abcd_idle)
            {
                start_call_disconnected(uc, dtmfr2->call);
            }
            else if (masked_abcd == dtmfr2->back_abcd_clear_back)
            {
                /* If they applied CLEAR BACK just before they saw our CLEAR FWD
                   this could legitimately occur. */
                start_far_disconnected(uc, dtmfr2->call);
                dtmfr2->r2_phase = R2_FWD_CLEAR_FWD_AFTER_CLEAR_BACK;
            }
            else
            {
                protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            }
            /*endif*/
            break;
        case R2_FWD_CLEAR_FWD_AFTER_CLEAR_BACK:
            if (masked_abcd == dtmfr2->abcd_idle)
                start_call_disconnected(uc, dtmfr2->call);
            else
                protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS);
            /*endif*/
            break;
        default:
            protocol_error_recovery(uc, ch, DTMFR2_PROTOCOL_FAIL_INVALID_STATE);
            break;
        }
        /*endswitch*/
    }
    /*endif*/
}
/*- End of function --------------------------------------------------------*/

static uc_event_t *check_event(uc_t *uc)
{
    uint8_t buf[1024];
    int16_t pcm_buf[1024];
    int res;
    int len;
    int xlen;
    int maxlen;
    int i;
    int x;
    int changes;
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;
    int ch;
    ZT_SPANINFO zi;

    ch = 0;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    for (;;)
    {
        //x = ZT_IOMUX_SIGEVENT | ZT_IOMUX_READ | ZT_IOMUX_WRITE | ZT_IOMUX_WRITEEMPTY | ZT_IOMUX_NOWAIT;
        x = ZT_IOMUX_SIGEVENT | ZT_IOMUX_READ | ZT_IOMUX_NOWAIT;
        switch (uc->chan[ch].active_tx)
        {
        case ACTIVE_SPEECH:
            if (uc->chan[ch].channel_write)
                x |= ZT_IOMUX_WRITE;
            /*endif*/
            break;
        case ACTIVE_DTMF:
            x |= ZT_IOMUX_WRITE;
            break;
        case ACTIVE_SUPER_TONE:
            x |= ZT_IOMUX_WRITE;
            break;
        }
        /*endif*/
        if ((res = ioctl(uc->chan[ch].handle, ZT_IOMUX, &x)))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            break;
        }
        /*endif*/
        if (x == 0)
            break;
        /*endif*/
        if ((x & ZT_IOMUX_SIGEVENT))
        {
            //uc_log(uc, UC_LOG_FLOW, "Event\n");
            res = ioctl(uc->chan[ch].handle, ZT_GETEVENT, &x);
            if (res == 0  &&  x != 0)
            {
                switch (x)
                {
                case ZT_EVENT_BITSCHANGED:
                    abcd_signaling_event(uc, ch);
                    break;
                case ZT_EVENT_ALARM:
                case ZT_EVENT_NOALARM:
                    memset(&zi, 0, sizeof(zi));
                    zi.spanno = 0;
                    if ((res = ioctl(uc->chan[ch].handle, ZT_SPANSTAT, &zi)) < 0)
                    {
                        if (uc->chan[ch].channel_error)
                            uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
                        /*endif*/
                        break;
                    }
                    /*endif*/
                    changes = dtmfr2->alarms ^ zi.alarms;
                    ev.alarm.e = UC_EVENT_ALARM;
                    ev.alarm.channel = ch;
                    ev.alarm.raised = changes & zi.alarms;
                    ev.alarm.cleared = changes & ~zi.alarms;
                    dtmfr2->alarms = zi.alarms;
                    uc_report_event(uc, &ev, sizeof(ev));
                    break;
                default:
                    if (uc->chan[ch].channel_error)
                        uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
                    /*endif*/
                    break;
                }
                /*endswitch*/
            }
            /*endif*/
            continue;
        }
        /*endif*/
        if ((x & ZT_IOMUX_READ))
        {
            //uc_log(uc, UC_LOG_FLOW, "Read - %d\n", uc->chan[ch].handle);
            if ((res = read(uc->chan[ch].handle, buf, sizeof(buf))) < 0)
            {
                if (errno == ELAST)
                {
                    /* This means there is some out of band stuff to deal with - alarms,
                       CAS bits signaling changes, etc. */
                }
                else if (errno != EAGAIN)
                {
                    if (uc->chan[ch].channel_error)
                        uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
                    /*endif*/
                }
                /*endif*/
                continue;
            }
            /*endif*/
            switch (uc->chan[ch].active_rx)
            {
            case ACTIVE_DTMF:
                /* Do the tone detection */
                switch (uc->chan[ch].api_codec)
                {
                case UC_CODEC_LINEAR16:
#if defined(AUDIO_LOG)
                    if (dtmfr2->audio_rx_log > 0)
                        write(dtmfr2->audio_rx_log, buf, res);
                    /*endif*/
#endif
                    dtmf_rx(&dtmfr2->dtmf_rx_state, (int16_t *) buf, res >> 1);
                    break;
                case UC_CODEC_ALAW:
#if defined(AUDIO_LOG)
                    if (dtmfr2->audio_rx_log > 0)
                        write(dtmfr2->audio_rx_log, buf, res);
                    /*endif*/
#endif
                    for (i = 0;  i < res;  i++)
                        pcm_buf[i] = alaw_to_linear(buf[i]);
                    /*endfor*/
                    dtmf_rx(&dtmfr2->dtmf_rx_state, pcm_buf, res);
                    break;
                case UC_CODEC_ULAW:
#if defined(AUDIO_LOG)
                    if (dtmfr2->audio_rx_log > 0)
                        write(dtmfr2->audio_rx_log, buf, res);
                    /*endif*/
#endif
                    for (i = 0;  i < res;  i++)
                        pcm_buf[i] = ulaw_to_linear(buf[i]);
                    /*endfor*/
                    dtmf_rx(&dtmfr2->dtmf_rx_state, pcm_buf, res);
                    break;
                default:
                    uc_log(uc, UC_LOG_WARNING, "Unknown codec %d\n", uc->chan[ch].api_codec);
                    break;
                }
                /*endswitch*/
                break;
            case ACTIVE_SPEECH:
                if (uc->chan[ch].channel_read)
                {
                    if (uc->chan[ch].channel_read_codec)
                        res = uc->chan[ch].channel_read_codec(uc, ch, uc->chan[ch].channel_read_codec_user_data, buf, res);
                    /*endif*/
                    uc->chan[ch].channel_read(uc, ch, uc->chan[ch].channel_read_user_data, buf, res);
                }
                /*endif*/
                break;
            }
            /*endswitch*/
            continue;
        }
        /*endif*/
        if ((x & ZT_IOMUX_WRITE))
        {
            //uc_log(uc, UC_LOG_FLOW, "Write\n");
            switch (uc->chan[ch].active_tx)
            {
            case ACTIVE_DTMF:
            case ACTIVE_SUPER_TONE:
                maxlen = uc->chan[ch].audio_bufsize;
                if (uc->chan[ch].active_tx == ACTIVE_DTMF)
                    len = dtmf_tx(&dtmfr2->dtmf_tx_state, pcm_buf, maxlen);
                else
                    len = super_tone_tx(&dtmfr2->super_tone_tx_state, pcm_buf, maxlen);
                /*endif*/
                switch (uc->chan[ch].api_codec)
                {
                case UC_CODEC_LINEAR16:
                    len *= 2;
#if defined(AUDIO_LOG)
                    if (dtmfr2->audio_tx_log > 0)
                        write(dtmfr2->audio_tx_log, pcm_buf, len);
                    /*endif*/
#endif
                    xlen = write(uc->chan[ch].handle, pcm_buf, len);
                    break;
                case UC_CODEC_ALAW:
                    for (i = 0;  i < len;  i++)
                        buf[i] = linear_to_alaw(pcm_buf[i]);
                    /*endfor*/
#if defined(AUDIO_LOG)
                    if (dtmfr2->audio_tx_log > 0)
                        write(dtmfr2->audio_tx_log, buf, len);
                    /*endif*/
#endif
                    xlen = write(uc->chan[ch].handle, buf, len);
                    break;
                case UC_CODEC_ULAW:
                    for (i = 0;  i < len;  i++)
                        buf[i] = linear_to_ulaw(pcm_buf[i]);
                    /*endfor*/
#if defined(AUDIO_LOG)
                    if (dtmfr2->audio_tx_log > 0)
                        write(dtmfr2->audio_tx_log, buf, len);
                    /*endif*/
#endif
                    xlen = write(uc->chan[ch].handle, buf, len);
                    break;
                default:
                    uc_log(uc, UC_LOG_WARNING, "Unknown codec %d\n", uc->chan[ch].api_codec);
                    break;
                }
                /*endswitch*/
                if (xlen != len)
                {
                    if (uc->chan[ch].channel_error)
                        uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
                    /*endif*/
                }
                /*endif*/
                break;
            case ACTIVE_SPEECH:
                if (uc->chan[ch].channel_write)
                {
                    maxlen = uc->chan[ch].audio_bufsize;
                    len = uc->chan[ch].channel_write(uc, ch, uc->chan[ch].channel_write_user_data, buf, maxlen);
                    if (len > 0)
                    {
                        if (uc->chan[ch].channel_write_codec)
                            len = uc->chan[ch].channel_write_codec(uc, ch, uc->chan[ch].channel_write_codec_user_data, buf, len);
                        /*endif*/
                        xlen = write(uc->chan[ch].handle, buf, len);
                        if (xlen != len)
                        {
                            if (uc->chan[ch].channel_error)
                                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
                            /*endif*/
                        }
                        /*endif*/
                    }
                    /*endif*/
                }
                /*endif*/
                break;
            }
            /*endswitch*/
            continue;
        }
        /*endif*/
        if ((x & ZT_IOMUX_WRITEEMPTY))
        {
            //uc_log(uc, UC_LOG_FLOW, "Write empty\n");
            continue;
        }
        /*endif*/
    }
    /*endfor*/
    return NULL;
}
/*- End of function --------------------------------------------------------*/

static void load_r2_parameter_set(dtmfr2_signaling_state_t *dtmfr2, int protocol, int outgoing_calls_allowed)
{
    dtmfr2->outgoing_calls_allowed = outgoing_calls_allowed;

    /* Set the standard values for everything first */
    dtmf_tx_init(&dtmfr2->dtmf_tx_state);
    dtmf_rx_init(&dtmfr2->dtmf_rx_state, handle_rx_dtmf, dtmfr2);

    dtmfr2->abcd_mask = 0xC;
    dtmfr2->base_abcd_bits = 0x1;

    dtmfr2->abcd_idle = 0x8;
    dtmfr2->abcd_blocking = 0xC;

    dtmfr2->fwd_abcd_seize = 0x0;
    dtmfr2->fwd_abcd_clear_fwd = 0x8;

    dtmfr2->back_abcd_seize_ack = 0xC;
    dtmfr2->back_abcd_answer = 0x4;
    dtmfr2->back_abcd_clear_back = 0xC;

    dtmfr2->abcd_billing_pulse = 0;
    dtmfr2->billing_pulse_duration = 0;

    dtmfr2->dnis_max_rx_digits = 7;
    dtmfr2->ani_max_rx_digits = UC_MAXPHONENUMLEN;

    dtmfr2->max_wait_for_group_b_signal = DEFAULT_MAX_WAIT_FOR_GROUP_B_SIGNAL;
    dtmfr2->max_await_answer = DEFAULT_MAX_AWAIT_ANSWER;

    dtmfr2->blocking_release_time = BLOCKING_RELEASE_TIME;
    dtmfr2->answer_guard_time = ANSWER_GUARD_TIME;
    dtmfr2->release_guard_time = 0;
    dtmfr2->inbound_cas_persistence_check = 20;
    dtmfr2->outbound_cas_persistence_check = 20;
    dtmfr2->max_seize_ack_wait = DEFAULT_MAX_SEIZE_ACK_WAIT;
    dtmfr2->clear_back_persistence_check = 0;
}
/*- End of function --------------------------------------------------------*/

static int make_call(uc_t *uc, uc_makecall_t *makecall)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;
    int ch;
    uc_call_t *call;
    const uc_callparms_t *callparms;
    
    ch = 0;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Make call\n");

    if (!dtmfr2->outgoing_calls_allowed)
        return  UC_RET_BLOCKED;
    /*endif*/

    if (uc->chan[ch].state == UC_STATE_BLOCKED)
        return  UC_RET_BLOCKED;
    /*endif*/
    if (uc->chan[ch].state != UC_STATE_IDLE)
        return  UC_RET_BAD_STATE;
    /*endif*/

    if ((call = uc_createcall(uc, 0)) == NULL)
        return UC_RET_MEMORY_ERROR;
    /*endif*/
    makecall->crn = call->crn;
    callparms = makecall->callparms;

    dtmfr2->call = call;
    strcpy(call->callparms.destination_number, callparms->destination_number);
    dtmfr2->dnis_len = strlen(call->callparms.destination_number);
    strcpy(call->callparms.originating_number, callparms->originating_number);
    dtmfr2->ani_len = strlen(call->callparms.originating_number);
    call->callparms.call_type = UC_CALL_TYPE_UNSPECIFIED;

    call->disconnect_reason = UC_CAUSE_UNSPECIFIED_CAUSE;

    call->incoming_call = FALSE;

    /* We don't need the detector ready until we start sending our first tone.
       Starting it early just makes us more vulnerable to false detection on
       whatever might be in the speech path. */
    if (set_abcd_signal(uc, ch, dtmfr2->fwd_abcd_seize))
        return  UC_RET_OK;
    /*endif*/
    dtmfr2->seize_ack_timer =
        uc_schedule_event(uc, dtmfr2->max_seize_ack_wait, seize_ack_wait_expired, (void *) (intptr_t) ch);
    
    call->chan = ch;
    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_DIALING;
    /*endif*/
    call->state = UC_STATE_DIALING;
    dtmfr2->r2_phase = R2_FWD_SEIZE;

    /* Once we SEIZE we must not clear until at least the SEIZE ACK has been
       received. Failure to do this is a protocol error. If the caller clears
       before SEIZE ACK, we must wait for SEIZE ACK before going to CLEAR FWD. */

    ev.e = UC_EVENT_DIALING;
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int send_more_info(uc_t *uc, uc_call_t *call, uc_sendmoreinfo_t *info)
{
    uc_log(uc, UC_LOG_FLOW, "Send more info\n");
    return  UC_RET_UNSUPPORTED;
}
/*- End of function --------------------------------------------------------*/

static int call_ack(uc_t *uc, uc_call_t *call, uc_callack_t *callack)
{
    uc_log(uc, UC_LOG_FLOW, "Call ack\n");
    return  UC_RET_UNSUPPORTED;
}
/*- End of function --------------------------------------------------------*/

static int accept_call(uc_t *uc, uc_call_t *call)
{
    dtmfr2_signaling_state_t *dtmfr2;
    int ch;

    ch = 0;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Accept call\n");
    if (call == NULL)
        return  UC_RET_BAD_STATE;
    /*endif*/
    if (call->state != UC_STATE_OFFERED)
        return  UC_RET_BAD_STATE;
    /*endif*/
    if (dtmfr2->call->callparms.call_type == UC_CALL_TYPE_NO_CHARGE_CALL)
    {
        dtmfr2->r2_state = R2_SENT_ACCEPTED_NO_CHARGE;
    }
    else
    {
        dtmfr2->r2_state = R2_SENT_ACCEPTED_CHARGE;
    }
    /*endif*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int answer_call(uc_t *uc, uc_call_t *call)
{
    dtmfr2_signaling_state_t *dtmfr2;
    int ch;
    
    ch = 0;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Answer call\n");
    if (call == NULL)
        return  UC_RET_BAD_STATE;
    /*endif*/
    switch (call->state)
    {
    case UC_STATE_OFFERED:
        /* We skipped the accept call step */
        /* Send the signal, indicating line free, charge or free. */
        if (dtmfr2->call->callparms.call_type == UC_CALL_TYPE_NO_CHARGE_CALL)
        {
            dtmfr2->r2_state = R2_SENT_ACCEPTED_NO_CHARGE;
        }
        else
        {
            dtmfr2->r2_state = R2_SENT_ACCEPTED_CHARGE;
        }
        /*endif*/
        break;
    case UC_STATE_ACCEPTED:
        /* Go directly to the answered state */
        if (dtmfr2->answer_guard_timer < 0)
        {
            select_active_rxtx(uc, ACTIVE_SPEECH);
            if (set_abcd_signal(uc, ch, dtmfr2->back_abcd_answer))
                return  UC_RET_OK;
            /*endif*/
            dtmfr2->r2_phase = R2_BACK_ANSWER;
            start_connected(uc, dtmfr2->call);
        }
        /*endif*/
        call->state = UC_STATE_CONNECTED;
        uc->chan[call->chan].state = UC_STATE_CONNECTED;
        break;
    default:
        return  UC_RET_BAD_STATE;
    }
    /*endif*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int drop_call(uc_t *uc, uc_call_t *call, int cause)
{
    dtmfr2_signaling_state_t *dtmfr2;
    int ch;
    
    ch = 0;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc,
           UC_LOG_FLOW, "Drop call(cause=%s [%d])\n",
           uc_cause2str(cause),
           cause);
    if (call == NULL)
        return  UC_RET_BAD_STATE;
    /*endif*/
    /* TODO: Check for further state conflicts */
    if (call->state == UC_STATE_NULL
        ||
        call->state == UC_STATE_IDLE
        ||
        call->state == UC_STATE_CALL_DISCONNECTED)
    {
        return  UC_RET_BAD_STATE;
    }
    /*endif*/
    if (call->incoming_call)
    {
        switch (call->state)
        {
        case UC_STATE_OFFERED:
            /* When a call has been offered dropping it requires we give a
               reason first. */
            switch (cause)
            {
            case UC_CAUSE_USER_BUSY:
                dtmfr2->r2_state = R2_SENT_DEST_BUSY;
                break;
            case UC_CAUSE_NETWORK_CONGESTION:
                dtmfr2->r2_state = R2_SENT_NETWORK_CONGESTION;
                break;
            case UC_CAUSE_UNASSIGNED_NUMBER:
                dtmfr2->r2_state = R2_SENT_UNASSIGNED_NUMBER;
                break;
            case UC_CAUSE_DEST_OUT_OF_ORDER:
            default:
                dtmfr2->r2_state = R2_SENT_DEST_OUT_OF_ORDER;
                break;
            }
            /*endswitch*/
            break;
        default:
            if (dtmfr2->r2_phase == R2_BACK_CLEAR_FWD)
            {
                /* Go to the call disconnected state, but do nothing else.
                   release call will initiate the final release. */
                start_call_disconnected(uc, call);
            }
            else
            {
                if (set_abcd_signal(uc, ch, dtmfr2->back_abcd_clear_back))
                    return  UC_RET_OK;
                /*endif*/
                dtmfr2->r2_phase = R2_BACK_CLEAR_BACK;
                if (dtmfr2->play_disconnect_tone)
                {
                    super_tone_tx_init(&dtmfr2->super_tone_tx_state, dtmfr2->super_tones->tone[ST_TYPE_DISCONNECTED]);
                    select_active_rxtx(uc, ACTIVE_SUPER_TONE);
                }
                /*endif*/
            }
            /*endif*/
            break;
        }
        /*endswitch*/
    }
    else
    {
        if (dtmfr2->await_answer_timer >= 0)
        {
            uc_schedule_del(uc, dtmfr2->await_answer_timer);
            dtmfr2->await_answer_timer = -1;
        }
        /*endif*/
        switch (call->state)
        {
        case UC_STATE_DIALING:
            if (dtmfr2->r2_phase == R2_FWD_SEIZE)
            {
                /* This must wait until SEIZE ACK either arrives or we time
                   out before we take any further action with the CAS bits. */
                dtmfr2->r2_phase = R2_FWD_CLEAR_FWD_BEFORE_SEIZE_ACK;
            }
            else
            {
                if (start_clear_fwd(uc))
                    return  UC_RET_BAD_DEVICE;
                /*endif*/
                dtmfr2->r2_phase = R2_FWD_CLEAR_FWD_BEFORE_ANSWER;
            }
            /*endif*/
            break;
        case UC_STATE_FAR_DISCONNECTED:
            if (start_clear_fwd(uc))
                return  UC_RET_BAD_DEVICE;
            /*endif*/
            dtmfr2->r2_phase = R2_FWD_CLEAR_FWD_AFTER_CLEAR_BACK;
            break;
        case UC_STATE_CONNECTED:
        default:
            if (start_clear_fwd(uc))
                return  UC_RET_BAD_DEVICE;
            /*endif*/
            dtmfr2->r2_phase = R2_FWD_CLEAR_FWD_AFTER_ANSWER;
            if (dtmfr2->play_disconnect_tone)
            {
                super_tone_tx_init(&dtmfr2->super_tone_tx_state, dtmfr2->super_tones->tone[ST_TYPE_DISCONNECTED]);
                select_active_rxtx(uc, ACTIVE_SUPER_TONE);
            }
            /*endif*/
            break;
        }
        /*endswitch*/
    }
    /*endif*/
    return UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int release_call(uc_t *uc, uc_call_t *call)
{
    dtmfr2_signaling_state_t *dtmfr2;
    int ch;
    
    ch = 0;    
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Release call\n");
    if (call == NULL)
        return  UC_RET_BAD_STATE;
    /*endif*/
    if (call->state != UC_STATE_CALL_DISCONNECTED)
        return  UC_RET_BAD_STATE;
    /*endif*/
    if (call->incoming_call)
    {
        if (set_abcd_signal(uc, ch, dtmfr2->abcd_idle))
            return  UC_RET_OK;
        /*endif*/
        select_active_rxtx(uc, ACTIVE_SPEECH);
        dtmfr2->r2_phase = R2_BACK_RELEASE_GUARD;

        /* Ensure a minimum period of release guard for the other end to be 
           sure of seeing it before we can seize for an outgoing call. A seize
           during this timeout will cause the seize logic to prematurely
           terminate this timer, and all will be well. */
        if (dtmfr2->release_guard_time == 0)
            dtmfr2->release_guard_time = 20;
        /*endif*/
        dtmfr2->release_guard_timer =
            uc_schedule_event(uc, dtmfr2->release_guard_time, release_guard_expired, (void *) call);
    }
    else
    {
        /* There is nothing to do here. The call has completely cleared already. */
        start_idle(uc, call);
    }
    /*endif*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int unblock(uc_t *uc, int ch)
{
    dtmfr2_signaling_state_t *dtmfr2;
    
    /* ch = -1 means all voice channels */
    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Unblock\n");
    set_abcd_signal(uc, 0, dtmfr2->abcd_idle);
    /* Start a timeout to ensure we do not seize the trunk too agressively.
       That is, before the far end has a chance to recognise the UNBLOCKED
       condition. */
    dtmfr2->local_unblocking_timer =
        uc_schedule_event(uc, dtmfr2->blocking_release_time, local_unblocking_expired, (void *) (intptr_t) ch);
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int block(uc_t *uc, int ch)
{
    dtmfr2_signaling_state_t *dtmfr2;
    uc_event_t ev;

    /* ch = -1 means all voice channels */
    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Block\n");
    dtmfr2->local_blocked = TRUE;
    set_abcd_signal(uc, 0, dtmfr2->abcd_blocking);
    ev.e = UC_EVENT_LOCALBLOCKED;
    ev.gen.channel = ch;
    ev.gen.crn = 0;
    ev.gen.call = NULL;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int call_control(uc_t *uc, int op, uc_call_t *call, void *data)
{
    dtmfr2_signaling_state_t *dtmfr2;
    int ch;

    ch = 0;
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Call control(%d)\n", op);
    switch (op)
    {
    case UC_OP_MAKECALL:
        return make_call(uc, (uc_makecall_t *) data);
    case UC_OP_SENDMOREINFO:
        return send_more_info(uc, call, (uc_sendmoreinfo_t *) data);
    case UC_OP_CALLACK:
        return call_ack(uc, call, (uc_callack_t *) data);
    case UC_OP_ACCEPTCALL:
        return accept_call(uc, call);
    case UC_OP_ANSWERCALL:
        return answer_call(uc, call);
    case UC_OP_DROPCALL:
        return drop_call(uc, call, (intptr_t) data);
    case UC_OP_RELEASECALL:
        return release_call(uc, call);
    case UC_OP_UNBLOCK:
        return unblock(uc, (intptr_t) data);
    case UC_OP_BLOCK:
        return block(uc, (intptr_t) data);
    }
    /*endswitch*/
    return  UC_RET_UNSUPPORTED;
}
/*- End of function --------------------------------------------------------*/

static uc_t *create_new(int fd, const char *variant, int mode, int outgoing_calls_allowed)
{
    uc_t *uc;
    dtmfr2_signaling_state_t *dtmfr2;
    uc_chan_t *chans;
    struct zt_params p;
    struct zt_bufferinfo b;
    int variation;
    int ch;
    int value;
    int len;
    const char *s;
    char *t;
    
    if ((variation = protocol_variant(variant)) < 0)
    {
        uc_error("Protocol variant '%s' unknown\n", variant);
        return  NULL;
    }
    /*endif*/
    if (ioctl(fd, ZT_GET_PARAMS, &p))
    {
        uc_error("Unable to get channel parameters: %s\n", strerror(errno));
        return  NULL;
    }
    /*endif*/
    if (p.sigtype != ZT_SIG_CAS)
    {
        uc_error("In signaling mode %d, instead of CAS.\n", p.sigtype);
        return  NULL;
    }
    /*endif*/
    if ((uc = malloc(sizeof(*uc))) == NULL)
    {
        uc_error("Failed to allocate buffers.\n");
        return  NULL;
    }
    /*endif*/
    memset(uc, '\0', sizeof(*uc));
    uc->protocol_variant = variation;
    uc->protocol_mode = mode;

    if ((dtmfr2 = malloc(sizeof(*dtmfr2))) == NULL)
    {
        uc_error("Failed to allocate buffers.\n");
        free(uc);
        return  NULL;
    }
    /*endif*/
    memset(dtmfr2, '\0', sizeof(*dtmfr2));
    uc->signaling = dtmfr2;

    dtmfr2->far_unblocking_timer =
    dtmfr2->local_unblocking_timer =
    dtmfr2->seize_ack_timer =
    dtmfr2->release_guard_timer =
    dtmfr2->answer_guard_timer =
    dtmfr2->await_answer_timer =
    dtmfr2->clear_back_persistence_timer =
    dtmfr2->billing_pulse_timer = -1;

    if ((chans = malloc(sizeof(*chans))) == NULL)
    {
        uc_error("Failed to allocate buffers.\n");
        free(dtmfr2);
        free(uc);
        return  NULL;
    }
    /*endif*/
    memset(chans, '\0', sizeof(*chans));
    uc->num_channels = 1;
    uc->chan = chans;
    ch = 0;
    uc->chan[ch].handle = fd;
    uc->chan[ch].state = UC_STATE_BLOCKED;
    uc->chan[ch].logging_level = 0;
    uc->chan[ch].active_rx =
    uc->chan[ch].active_tx = ACTIVE_SPEECH;
    uc->chan[ch].native_law =
    uc->chan[ch].current_law = p.curlaw;
    uc->chan[ch].native_codec =
    uc->chan[ch].api_codec = (p.curlaw == ZT_LAW_ALAW)  ?  UC_CODEC_ALAW  :  (p.curlaw == ZT_LAW_MULAW)  ?  UC_CODEC_ULAW  :  UC_CODEC_UNKNOWN;
#if defined(AUDIO_LOG)
    dtmfr2->audio_rx_log =
    dtmfr2->audio_tx_log = -1;
#endif
#if defined(AUDIO_LOG)
    {
        char fred[100];
        int yyy;
        static int xxx = 1;

        yyy = xxx++;
        sprintf(fred, "channel-tx-%d.au", yyy);
        dtmfr2->audio_tx_log = open(fred, O_CREAT | O_TRUNC | O_RDWR, 0666);
        if (dtmfr2->audio_tx_log < 0)
        {
            printf("Failed to open '%s' - %d\n", fred, errno);
            exit(2);
        }
        else
        {
            printf("Opened '%s' - %d\n", fred, dtmfr2->audio_tx_log);
        }
        /*endif*/
        sprintf(fred, "channel-rx-%d.au", yyy);
        dtmfr2->audio_rx_log = open(fred, O_CREAT | O_TRUNC | O_RDWR, 0666);
        if (dtmfr2->audio_rx_log < 0)
        {
            printf("Failed to open '%s' - %d\n", fred, errno);
            exit(2);
        }
        else
        {
            printf("Opened '%s' - %d\n", fred, dtmfr2->audio_rx_log);
        }
        /*endif*/
    }
#endif

    if (ioctl(fd, ZT_GET_BUFINFO, &b))
    {
        free(chans);
        free(dtmfr2);
        free(uc);
        uc_error("Unable to get channel buffer info: %s\n", strerror(errno));
        return  NULL;
    }
    /*endif*/
    uc->chan[ch].audio_bufsize = b.bufsize;

    load_r2_parameter_set(dtmfr2, variation, outgoing_calls_allowed);
    
    s = variant;
    if (variant_element(&s, 1))
    {
        if (sscanf(s, "%d", &value) == 1
            &&
            value >= 0  &&  value <= UC_MAXPHONENUMLEN)
        {
            dtmfr2->ani_max_rx_digits = value;
        }
        /*endif*/
    }
    /*endif*/
    s = variant;
    if (variant_element(&s, 2))
    {
        if (sscanf(s, "%d", &value) == 1
            &&
            value >= 0  &&  value <= UC_MAXPHONENUMLEN)
        {
            dtmfr2->dnis_max_rx_digits = value;
        }
        /*endif*/
    }
    /*endif*/
    s = variant;
    if (variant_element(&s, 3))
    {
        if (sscanf(s, "%d", &value) == 1)
        {
            dtmfr2->play_progress_tones = (value & 1)  ?  TRUE  :  FALSE;
            dtmfr2->play_disconnect_tone = (value & 2)  ?  TRUE  :  FALSE;
            dtmfr2->play_ringback_tone = (value & 4)  ?  TRUE  :  FALSE;
        }
        /*endif*/
    }
    /*endif*/
    
    if (dtmfr2->play_progress_tones  ||  dtmfr2->play_disconnect_tone  ||  dtmfr2->play_ringback_tone)
    {
        s = variant;
        len = variant_element(&s, 0);
        t = malloc(len + 1);
        strncpy(t, s, len);
        t[len] = '\0';
        dtmfr2->super_tones = get_supervisory_tone_set(t);
        free(t);
        if (dtmfr2->super_tones == NULL)
        {
            free(chans);
            free(dtmfr2);
            free(uc);
            uc_error("Failed to load supervisory tones.\n");
            return  NULL;
        }
        /*endif*/
    }
    /*endif*/

    hash_InitHashTable(&uc->calls, HASH_ONE_WORD_KEYS);

    dtmfr2->current_abcd_in = 0xFF;
    if (set_abcd_signal(uc, ch, dtmfr2->abcd_blocking) == UC_RET_OK)
        abcd_signaling_event(uc, ch);
    /*endif*/
    return uc;
}
/*- End of function --------------------------------------------------------*/

static int channel_open(const char *variant, char *dev)
{
    struct zt_bufferinfo bi;
    struct zt_gains g;
    int x;
    int fd;
    int i;
    int chan;

    if ((chan = atoi(dev)) > 0)
    {
        if ((fd = open("/dev/zap/channel", O_RDWR | O_NONBLOCK)) < 0)
            return -1;
        /*endif*/
        if (ioctl(fd, ZT_SPECIFY, &chan))
        {
            x = errno;
            close(fd);
            errno = x;
            return -1;
        }
        /*endif*/
    }
    else
    {
        if ((fd = open(dev, O_RDWR | O_NONBLOCK)) < 0)
            return -1;
        /*endif*/
    }
    /*endif*/
    if (ioctl(fd, ZT_GET_BUFINFO, &bi) < 0)
    {
        x = errno;
        close(fd);
        errno = x;
        return -1;
    }
    /*endif*/
    bi.txbufpolicy = ZT_POLICY_IMMEDIATE;
    bi.rxbufpolicy = ZT_POLICY_IMMEDIATE;
    bi.numbufs = 4;
    bi.bufsize = 160;
    if (ioctl(fd, ZT_SET_BUFINFO, &bi) < 0)
    {
        x = errno;
        close(fd);
        errno = x;
        return -1;
    }
    /*endif*/
    if (ioctl(fd, ZT_CHANNO, &chan))
    {
        x = errno;
        close(fd);
        errno = x;
        return -1;
    }
    /*endif*/
    /* Set default gains */
    g.chan = 0;
    for (i = 0;  i < 256;  i++)
    {
        g.rxgain[i] = i;
        g.txgain[i] = i;
    }
    /*endfor*/
    if (ioctl(fd, ZT_SETGAINS, &g) < 0)
    {
        x = errno;
        close(fd);
        errno = x;
        return -1;
    }
    /*endif*/
    return fd;
}
/*- End of function --------------------------------------------------------*/

static int channel_close(int fd)
{
    close(fd);
    return 0;
}
/*- End of function --------------------------------------------------------*/

static char *error_message(uc_t *uc, int code)
{
    switch (code)
    {
    case DTMFR2_PROTOCOL_FAIL_UNEXPECTED_DTMF_SIGNAL:
        return "Unexpected DTMF signal";
    case DTMFR2_PROTOCOL_FAIL_UNEXPECTED_CAS:
        return "Unexpected CAS bit pattern";
    case DTMFR2_PROTOCOL_FAIL_INVALID_STATE:
        return "Invalid state";
    case DTMFR2_PROTOCOL_FAIL_SET_CAS_FAILURE:
        return "Failed to set CAS signaling bits";
    case DTMFR2_PROTOCOL_FAIL_SEIZE_ACK_TIMEOUT:
        return "Seize ack timed out";
    case DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR:
        return "Device I/O error";
    }
    /*endswitch*/
    return "Undefined";
}
/*- End of function --------------------------------------------------------*/

static int get_device_handles(uc_t *uc, int ch, int *fe, int *be)
{
    if (fe)
        *fe = uc->chan[ch].handle;
    /*endif*/
    if (be)
        *be = uc->chan[ch].handle;
    /*endif*/
    return UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int channel_set_api_codec(uc_t *uc, int ch, int codec)
{
    int linear;
    int law;

    //uc_log(uc, UC_LOG_FLOW, "Channel set API codec - %d\n", codec);
    switch (codec)
    {
    case UC_CODEC_DEFAULT:
        linear = 0;
        if (ioctl(uc->chan[ch].handle, ZT_SETLINEAR, &linear))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        law = uc->chan[ch].native_law;
        if (law == ZT_LAW_ALAW)
            codec = UC_CODEC_ALAW;
        else if (law == ZT_LAW_MULAW)
            codec = UC_CODEC_ULAW;
        /*endif*/
        break;
    case UC_CODEC_LINEAR16:
        linear = 1;
        if (ioctl(uc->chan[ch].handle, ZT_SETLINEAR, &linear))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        law = uc->chan[ch].native_law;
        break;
    case UC_CODEC_ALAW:
        linear = 0;
        if (ioctl(uc->chan[ch].handle, ZT_SETLINEAR, &linear))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        law = ZT_LAW_ALAW;
        break;
    case UC_CODEC_ULAW:
        linear = 0;
        if (ioctl(uc->chan[ch].handle, ZT_SETLINEAR, &linear))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        law = ZT_LAW_MULAW;
        break;
    default:
        return  UC_RET_UNSUPPORTED;
    }
    /*endswitch*/
    if (uc->chan[ch].current_law != law)
    {
        uc->chan[ch].current_law = law;
        if (ioctl(uc->chan[ch].handle, ZT_SETLAW, &law))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return UC_RET_DEVICE_ERROR;
        }
        /*endif*/
    }
    /*endif*/
    uc->chan[ch].api_codec = codec;
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int channel_write(uc_t *uc, int ch, uint8_t *buf, int len)
{
    int xlen;
    int i;

    if (uc->chan[ch].active_tx != ACTIVE_SPEECH)
        return len;
    /*endif*/
    if (uc->chan[ch].channel_write_codec)
        len = uc->chan[ch].channel_write_codec(uc, ch, uc->chan[ch].channel_write_codec_user_data, buf, len);
    /*endif*/
    /* NASTY BODGE ALERT:
       EAGAIN means many things, including write buffer full. We have no way to
       know what actually caused it. We need to limit the number of write tries,
       otherwise buffer full can foul up the whole machine with a storm of
       retries. Sleeps between tries are possible, but that is not a cure.
       If we get a buffer full condition, something is writing too fast.
       Loosing some data is the only reasonable strategy. */
    for (i = 0;  i < 5;  i++)
    {
        xlen = write(uc->chan[ch].handle, buf, len);
        if (xlen >= len)
            break;
        /*endif*/
        if (xlen < 0)
        {
            if (errno == ELAST)
            {
                check_event(uc);
                continue;
            }
            /*endif*/
            if (errno == EAGAIN)
                continue;
            /*endif*/
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            break;
        }
        /*endif*/
        len -= xlen;
        buf += len;
    }
    /*endfor*/
    if (errno == EAGAIN)
        xlen = 1;
    /*endif*/
    return  xlen;
}
/*- End of function --------------------------------------------------------*/

static int channel_flush(uc_t *uc, int ch, int which)
{
    dtmfr2_signaling_state_t *dtmfr2;
    int x;

    /* ch = -1 means all voice channels */
    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Channel flush\n");
    if ((which & 1))
    {
        x = ZT_FLUSH_WRITE;
        if (ioctl(uc->chan[ch].handle, ZT_FLUSH, &x))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
    }
    /*endif*/
    if ((which & 2))
    {
        /* TODO: implement input flush */
    }
    /*endif*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

int channel_gains(uc_t *uc, int ch, float rxgain, float txgain)
{
    dtmfr2_signaling_state_t *dtmfr2;
    struct zt_gains g;
    float gain;
    int j;
    int k;

    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Channel gains\n");
    g.chan = 0;
    if (rxgain != 0.0)
    {
        /* Calculate linear value of rx gain */
        gain = pow(10.0, rxgain/20.0);
        if (uc->chan[ch].current_law == UC_CODEC_ALAW)
        {
            for (j = 0;  j < 256;  j++)
            {
                k = fsaturate(((float) alaw_to_linear(j))*gain);
                g.rxgain[j] = linear_to_alaw(k);
            }
            /*endfor*/
        }
        else
        {
            for (j = 0;  j < 256;  j++)
            {
                k = fsaturate(((float) ulaw_to_linear(j))*gain);
                g.rxgain[j] = linear_to_ulaw(k);
            }
            /*endfor*/
        }
        /*endif*/
    }
    else
    {
        for (j = 0;  j < 256;  j++)
            g.rxgain[j] = j;
        /*endfor*/
    }
    /*endif*/
    if (txgain != 0.0)
    {
        /* Calculate linear value of tx gain */
        gain = pow(10.0, txgain/20.0);
        if (uc->chan[ch].current_law == UC_CODEC_ALAW)
        {
            for (j = 0;  j < 256;  j++)
            {
                k = fsaturate(((float) alaw_to_linear(j))*gain);
                g.txgain[j] = linear_to_alaw(k);
            }
            /*endfor*/
        }
        else
        {
            for (j = 0;  j < 256;  j++)
            {
                k = fsaturate(((float) ulaw_to_linear(j))*gain);
                g.txgain[j] = linear_to_ulaw(k);
            }
            /*endfor*/
        }
        /*endif*/
    }
    else
    {
        for (j = 0;  j < 256;  j++)
            g.txgain[j] = j;
        /*endfor*/
    }
    /*endif*/
    if (ioctl(uc->chan[ch].handle, ZT_SETGAINS, &g))
    {
        if (uc->chan[ch].channel_error)
            uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
        /*endif*/
        return  UC_RET_DEVICE_ERROR;
    }
    /*endif*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

int channel_echo_cancel(uc_t *uc, int ch, int op)
{
    dtmfr2_signaling_state_t *dtmfr2;
    int x;

    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Channel echo cancel\n");
    switch (op)
    {
    case UC_ECHO_CANCEL_OFF:
    case UC_ECHO_CANCEL_ON:
        x = op;
        if (ioctl(uc->chan[ch].handle, ZT_ECHOCANCEL, &x) < 0)
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        break;
    case UC_ECHO_CANCEL_NOTRAINING:
    case UC_ECHO_CANCEL_TRAINING:
        x = op - UC_ECHO_CANCEL_NOTRAINING;
        if (ioctl(uc->chan[ch].handle, ZT_ECHOTRAIN, &x) < 0)
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        break;
    default:
        return UC_RET_BAD_PARAMETER;
    }
    /*endswitch*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

int channel_switching(uc_t *uc, int ch, int op, int dest, int parms)
{
    struct zt_confinfo zi;
    dtmfr2_signaling_state_t *dtmfr2;
    int x;

    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    dtmfr2 = (dtmfr2_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Channel switching\n");
    switch (op)
    {
    case UC_SWITCHING_FREE:
        memset(&zi, 0, sizeof(zi));
        if (ioctl(uc->chan[ch].handle, ZT_SETCONF, &zi))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        break;
    case UC_SWITCHING_UNMUTE:
    case UC_SWITCHING_MUTE:
        x = op - UC_SWITCHING_UNMUTE;
        if (ioctl(uc->chan[ch].handle, ZT_CONFMUTE, &x))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        break;
    case UC_SWITCHING_CONNECT:
        memset(&zi, 0, sizeof(zi));
        zi.confno = dest;
        zi.confmode = parms;
        if (ioctl(uc->chan[ch].handle, ZT_SETCONF, &zi))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, DTMFR2_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        break;
    default:
        return UC_RET_BAD_PARAMETER;
    }
    /*endswitch*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/
/*- End of file ------------------------------------------------------------*/
