/*
 * libdtmfr2 - a library for DTMF/R2 signaling on E1 lines
 *
 * dtmfr2.h
 *
 * Written by Steve Underwood <steveu@coppice.org>
 *
 * Copyright (C) 2002, 2003 Steve Underwood
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: dtmfr2.h,v 1.2 2005/12/10 14:48:48 steveu Exp $
 */

#if !defined(_DTMFR2_H_)
#define _DTMFR2_H_

/* Welcome to the weird and wonderful world of DTMF/R2 signaling */

enum
{
    /* Call states for R2 protocol inbound (backward) */
    /*! Sending SEIZE ACK in response to SEIZE */
    R2_BACK_SEIZE_ACK                           = 101,
    /*! Sending DTMF */
    R2_BACK_DTMF                                = 102,
    /*! Sending ANSWER signal, after receiving SIGII_1 tone off */
    R2_BACK_ANSWER                              = 105,
    /*! Sending CLEAR BACK after ANSWER */
    R2_BACK_CLEAR_BACK                          = 106,
    /*! Got CLEAR FWD before CLEAR BACK */
    R2_BACK_CLEAR_FWD                           = 107,
    /*! Sending timed RELEASE GUARD to prevent early re-seize from our side. */
    R2_BACK_RELEASE_GUARD                       = 108,

    /* Call states for R2 protocol outbound (forward) */
    /*! Sending SEIZE */
    R2_FWD_SEIZE                                = 201,
    /*! Proceeding with DTMF signals */
    R2_FWD_DTMF                                 = 202,
    /*! Waiting for CAS ANSWER after group B signal received */
    R2_FWD_AWAIT_ANSWER                         = 205,
    /*! CAS ANSWER has preceeded group B signal */
    R2_FWD_ANSWERED_BEFORE_GROUP_B_SIGNAL       = 206,
    /*! The call is fully answered (group B tone + CAS bits) */
    R2_FWD_ANSWERED                             = 207,
    /*! CLEAR BACK after ANSWER */
    R2_FWD_CLEAR_BACK                           = 208,
    /*! Want to send CLEAR FWD before receiving SEIZE ACK */
    R2_FWD_CLEAR_FWD_BEFORE_SEIZE_ACK           = 209,
    /*! Sending CLEAR FWD before receiving ANSWER */
    R2_FWD_CLEAR_FWD_BEFORE_ANSWER              = 210,
    /*! Sending CLEAR FWD after receiving ANSWER */
    R2_FWD_CLEAR_FWD_AFTER_ANSWER               = 211,
    /*! Sending CLEAR FWD after receiving CLEAR BACK. Waiting for RELEASE GUARD */
    R2_FWD_CLEAR_FWD_AFTER_CLEAR_BACK           = 212
} r2_phase_e;

/* Now some definitions for the states of the R2 protocol level state machine */
enum
{
    /* Group A states for R2 protocol inbound (backward). */
    R2_SENT_SEIZE_ACK                           = 301,

    /* Group B states for R2 protocol inbound (backward) */
    /*! Sent numbers complete and start ANSWER. Charge. */
    R2_SENT_ACCEPTED_CHARGE                     = 306,
    /*! Sent numbers complete and start ANSWER. No charge. */
    R2_SENT_ACCEPTED_NO_CHARGE                  = 307,
    /*! Sent numbers complete. Route congested. */
    R2_SENT_NETWORK_CONGESTION                  = 308,
    /*! Sent numbers complete. Destination number unassigned. */
    R2_SENT_UNASSIGNED_NUMBER                   = 309,
    /*! Sent numbers complete. Destination number busy. */
    R2_SENT_DEST_BUSY                           = 310,
    /*! Sent numbers complete. Destination number out of order. */
    R2_SENT_DEST_OUT_OF_ORDER                   = 311,

    /* Group I states for R2 protocol outbound (forward) */
    R2_SENT_DNIS                                = 402,
    R2_SENT_ANI                                 = 403,
    R2_SENT_END_OF_DNIS                         = 404,
    R2_SENT_END_OF_ANI                          = 405,
    R2_SENT_NOTHING                             = 406
} r2_state_e;

/* This structure should contain everything to control an R2 channel. */
struct dtmfr2_signaling_state_s
{
    /* Configuration parameters, to adapt to different countries. */
    /*! \brief The number of DNIS digits we expect for incoming calls. Some national
               variants do not handle variable length dialed digit strings well. In
               those cases it is important to set this correctly. For other variants,
               a large number is OK. */
    int dnis_max_rx_digits;
    /*! \brief The number of ANI digits we expect for incoming calls. */
    int ani_max_rx_digits;
    /*! \brief TRUE to play progress tones for busy, NU, congested, etc. */
    int play_progress_tones;
    /*! \brief TRUE to play the disconnect tone after the local end has disconnected. */
    int play_disconnect_tone;
    /*! \brief TRUE to play the ringback tone while waiting for answer. */
    int play_ringback_tone;
    /*! \brief */
    int blocking_release_time;
    /*! \brief */
    int answer_guard_time;
    /*! \brief Zero may be OK */
    int release_guard_time;
    /*! \brief 20ms should be OK */
    int inbound_cas_persistence_check;
    /*! \brief 20ms should be OK */
    int outbound_cas_persistence_check;
    /*! ITU spec says 200ms */
    int max_seize_ack_wait;
    /*! \brief The maximum time we will for the called party to answer, after the
               call setup is complete. 60s is common practice */
    int max_await_answer;
    /*! \brief The maximum time we will wait for the group B signal. 15s is common
               practice */
    int max_wait_for_group_b_signal;
    /*! \brief Clear back persistence check, in milliseconds. This can be zero for
               most purposes */
    int clear_back_persistence_check;
    /*! \brief All trunks are able to accept incoming calls. This flag is TRUE if the
               trunk is permitted to also make outgoing calls. */
    int outgoing_calls_allowed;

    /*! \brief Mask to pick out the R2 signaling bits from amongst the ABCD bits. */
    uint8_t abcd_mask;
    /*! \brief The value to use on transmission for the ABCD bits which are not the
               R2 signaling bits. */
    uint8_t base_abcd_bits;
    /*! \brief The idle pattern for the ABCD bits */
    uint8_t abcd_idle;
    /*! \brief The blocking pattern for the ABCD bits */
    uint8_t abcd_blocking;

    /*! \brief The seize pattern for the ABCD bits */
    uint8_t fwd_abcd_seize;
    /*! \brief The clear forward pattern for the ABCD bits */
    uint8_t fwd_abcd_clear_fwd;

    /*! \brief The seize acknowledge pattern for the ABCD bits */
    uint8_t back_abcd_seize_ack;
    /*! \brief The answer pattern for the ABCD bits */
    uint8_t back_abcd_answer;
    /*! \brief The clear back pattern for the ABCD bits */
    uint8_t back_abcd_clear_back;

    uint8_t abcd_billing_pulse;
    int billing_pulse_duration;

    /*! \brief The current call information. */
    uc_call_t *call;
    
    /*! \brief Alarms */
    int alarms;

    /*! \brief The current phase of the R2 protocol. */
    int r2_phase;
    /*! \brief The current compelled signalling state of the R2 protocol. */
    int r2_state;
    /*! \brief A flag used to differentiate calling party category requests from
               ANI digit requests. They use the same tone in many places. */
    int calling_party_category_exchanged;
    /*! \brief A flag which indicates all DNIS digits have been received. */
    int dnis_complete;
    /*! \brief The total length of the DNIS string for outgoing calls. A progressively
               incrementing count of the DNIS digits for incoming calls. */
    int dnis_len;
    /*! \brief The total length of the ANI string for outgoing calls. A progressively
               incrementing count of the ANI digits for incoming calls. */
    int ani_len;
    /*! \brief The DNIS digits sent so far for outgoing calls. Not used for incoming calls. */
    int dnis_cnt;
    /*! \brief The ANI digits sent so far for outgoing calls. Not used for incoming calls. */
    int ani_cnt;
    /*! \brief The ANI is restricted. */
    int ani_restricted;
    /*! \brief The current CAS ABCD pattern being received. */
    uint8_t current_abcd_in;
    /*! \brief Timer used to ensure we do not seize to aggressively as the far end unblocks. */
    int far_unblocking_timer;
    /*! \brief Timer used to ensure we do not seize to aggressively after we unblock. */
    int local_unblocking_timer;
    /*! \brief The seize acknowledge timeout, used to detect failure at the remote end. */
    int seize_ack_timer;
    /*! \brief The release guard timeout. */
    int release_guard_timer;
    /*! \brief Optional minimum delay between sending the group B signal and
        answering.
    
        If we answer too soon after sending the group B signal, the
        outgoing side might see the answer before its tone detector has
        found the group B signal. This timer allows some delay to be
        introduced, to prevent such overlap conditions. */
    int answer_guard_timer;
    /*! \brief A timer to handle the possibility that the called party never answers. */
    int await_answer_timer;
    /*! \brief Clear back persistence timer, used to differentiate between metering pulses
        and real clear back conditions. */
    int clear_back_persistence_timer;
    /*! \brief Timer for the duration of a billing pulse. */
    int billing_pulse_timer;
    /*! \brief TRUE is the far end is in the blocked state. */
    int far_blocked;
    /*! \brief TRUE is the local end is in the blocked state. */
    int local_blocked;
    /*! \brief The DTMF receiver context. */
    dtmf_rx_state_t dtmf_rx_state;
    /*! \brief The DTMF transmitter context. */
    dtmf_tx_state_t dtmf_tx_state;
    /*! \brief Supervisory tone generator context. */
    super_tone_tx_state_t super_tone_tx_state;
    /*! \brief Supervisory tone set. */
    super_tone_set_t *super_tones;

    /*! \brief Debug audio receive log file handle. */
    int audio_rx_log;
    /*! \brief Debug audio transmit log file handle. */
    int audio_tx_log;
};

#endif
